#! /usr/bin/perl 
######
#~
#~ KI für FreiesMagazin-Wettbewerb
#~
#~ Copyright (C) 2009 Robert Jackermeier
#~ 
#~ This program is free software; you can redistribute it and/or modify 
#~ it under the terms of the GNU General Public License as published by 
#~ the Free Software Foundation; either version 3 of the License, or (at 
#~ your option) any later version.
#~ 
#~ This program is distributed in the hope that it will be useful, 
#~ but WITHOUT ANY WARRANTY; without even the implied warranty of 
#~ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
#~ General Public License for more details.
#~ 
#~ You should have received a copy of the GNU General Public License 
#~ along with this program; if not, see <http://www.gnu.org/licenses/>.
#~ 
######

use strict;
use warnings;

## Dateien öffnen:
open PLAYER, '<', 'player.dat';
open OPPONENT, '<', 'opponent.dat';
open GAMEFIELD, '<', 'gamefield.dat';
open RESULT, '>', 'result.dat';

## Ein paar wichtige Variablen:
my %player_stats;
my %opponent_stats;
my @spielfeld;
my $zug;

## Spielerdaten einlesen:
($player_stats{'lifepoints'}, $player_stats{'shield'}, $player_stats{'red'}, $player_stats{'yellow'}, $player_stats{'green'}, $player_stats{'lilac'})  = <PLAYER>;
($opponent_stats{'lifepoints'}, $opponent_stats{'shield'}, $opponent_stats{'red'}, $opponent_stats{'yellow'}, $opponent_stats{'green'}, $opponent_stats{'lilac'})  = <OPPONENT>;
close PLAYER;
close OPPONENT;

## Spielfeld einlesen:
my $lines_count = 9;
foreach my $line (<GAMEFIELD>){
	map {push @{ $spielfeld[$lines_count] }, $_ } split " ", $line;
	$lines_count--;
}
close GAMEFIELD;

## Backup des Original-Spielfelds:
my $spielfeld = \@spielfeld;
my @spielfeld_alt;
foreach my $line (0..9) {
	map { push @{ $spielfeld_alt[$line]}, $_ } @{ $spielfeld[$line] };	
}

## Durch alle Zellen iterieren und mögliche Züge vorausberechnen:
my $punkte_maximum = 0;
my $punkte_current = 0;

for my $zeile (0..9) {
	for my $spalte (0..9) {
		
		$punkte_current = 0;
		
		## Horizontal tauschen (aktuelle Zelle nach rechts)
		if ($spalte < 9) {
			
			# Feld in Ursprungszustand versetzen:											
			@spielfeld = ();
			foreach my $line (0..9) {
				map { push @{ $spielfeld[$line]}, $_ } @{ $spielfeld_alt[$line] };	
			}
	 
			# Zellen tauschen:				
			($spielfeld->[$zeile]->[$spalte], $spielfeld->[$zeile]->[$spalte+1]) = ($spielfeld->[$zeile]->[$spalte+1], $spielfeld->[$zeile]->[$spalte]);
			
			# Punkte für Zug ermitteln:
			$punkte_current = get_points($spielfeld, \%player_stats, $spalte, $spalte+1, $zeile, $zeile); 		
			
			# Prüfen, ob Zug besser ist als die bisherigen:
			if ($punkte_current > $punkte_maximum) {
				$punkte_maximum = $punkte_current;
				$zug = "$spalte $zeile ". ($spalte+1) . " $zeile";
			}
		}
			
		$punkte_current = 0;

		# Vertikal tauschen (aktuelle Zelle nach oben)
		if ($zeile < 9) {
			
			# Feld in Ursprungszustand versetzen:	
			@spielfeld = ();
			foreach my $line (0..9) {
				map { push @{ $spielfeld[$line]}, $_ } @{ $spielfeld_alt[$line] };	
			}
			
			# Zellen tauschen:	
			($spielfeld->[$zeile]->[$spalte], $spielfeld->[$zeile+1]->[$spalte]) = ($spielfeld->[$zeile+1]->[$spalte], $spielfeld->[$zeile]->[$spalte]);

			# Punkte für Zug ermitteln:
			$punkte_current = get_points($spielfeld, \%player_stats, $spalte, $spalte, $zeile, $zeile+1); 		
			
			# Prüfen, ob Zug besser ist als die bisherigen:
			if ($punkte_current > $punkte_maximum) {
				$punkte_maximum = $punkte_current;
				$zug = "$spalte $zeile $spalte ". ($zeile+1) ;
			}
		}
	}
}

## Zug an Spiellogik übergeben:
print RESULT $zug;
close RESULT;

#### Subfunktionen:

## Punktzahl für einzelnen Zug berechnen:
sub get_points {
	## übergebene Variablen einlesen:
	my $spielfeld = shift;
	my $player_stats = shift;
	my ($x0, $x1, $y0, $y1) = @_;
	
	## Variablen initialisieren:
	my $points = 0;
	my $oldpoints = 0;
	my $zu_eliminierende_zellen;
	
	## Punktesystem:
	my %farbwerte = (	'R'=>'10',			# 15x Rot entspr. 10 Schaden
						'Y'=>'6',			# 15x Gelb entspr. 6 Schaden
						'G'=>'3',			# 15x Grün entspr. 3 Schaden
						'B'=>'15', 			# 15x Blau entspr. 15 verhindertem Schaden
						'L'=>'8',			# Wert hat sich in Tests bewährt
					 	'1'=>'15',			# 15x Bombe1 entspr. 15 Schaden
						'2'=>'30',			# 15x Bombe2 entspr. 30 Schaden
						'3'=>'45',			# 15x Bombe3 entspr. 45 Schaden
						'4'=>'60',			# 15x Bombe4 entspr. 60 Schaden
						'5'=>'75',			# 15x Bombe5 entspr. 75 Schaden
	);
	
	
	while (){
		
		$oldpoints = $points;
		
		## Zellen mit 3+-Kombinationen finden:
		$zu_eliminierende_zellen = find_combos($spielfeld, $x0, $x1, $y0, $y1);
		
		## Grenzen zurücksetzen:
		($x0, $x1, $y0, $y1) = undef;		
		
		## Werte der Zellen notieren
		for my $zeile (0..9) {

			for my $spalte (0..9) {
					
				## Falls Zelle zu einer Kombination gehört:
				if ($zu_eliminierende_zellen->[$zeile]->[$spalte] == 1){
					
					## Grenzen für Überprüfung des Feldes festlegen 
					# linke Grenze:
					if (not defined $x0 ) {								
						$x0 = $spalte ;
					}
					elsif ($spalte < $x0) { 							
						$x0 = $spalte ;
					}
					# rechte Grenze:										
					if ((defined $x1) and ($spalte > $x1)) {			
						$x1 = $spalte ;	
					}
					elsif (not defined $x1) {
						$x1 = $spalte;
					}
					# untere Grenze:
					$y0 = $zeile unless defined $y0;
					# obere Grenze:
					$y1 = 9;											
					
					
					## Blau nur wichtig, wenn Schild nicht voll:
					if (($spielfeld->[$zeile]->[$spalte] eq 'B') and ($player_stats->{'shield'} == 15)) {
						$points += 1; 
					}
					## Ansonsten Punktwert dieser Zelle zur Punktzahl des Zuges adieren:
					else {
						$points += $farbwerte{$spielfeld->[$zeile]->[$spalte]};
					}
				}
			}	
		}
		
		## falls keine Punkte dazugekommen sind, d.h. keine Kombinationen gefunden wurden, abbrechen:
		last if $oldpoints == $points;
		
		
		## Zellen eliminieren:
		for my $zeile (0..9) {

			for my $spalte (0..9) {
		
				## Spezialbehandlung für nachgerutschte Zellen:
				unless ($zu_eliminierende_zellen->[$zeile]->[$spalte]) {
					$zu_eliminierende_zellen->[$zeile]->[$spalte] = 0;
				}
				
				## Falls Zelle in einer Kombination...
				while ($zu_eliminierende_zellen->[$zeile]->[$spalte] == 1) {
					
					foreach my $y_wert ($zeile..9){
						
						## ...Zelle löschen und alle darüberliegenden Zellen um eins nach unten versetzen...
						if ($spielfeld->[$y_wert+1]->[$spalte]){
							$spielfeld->[$y_wert]->[$spalte] = $spielfeld->[$y_wert+1]->[$spalte];
							$zu_eliminierende_zellen->[$y_wert]->[$spalte] = $zu_eliminierende_zellen->[$y_wert+1]->[$spalte];
						}
						## ...und ganz oben neue Zellen mit Nullen einfügen (Hier könnte man alle Möglichkeiten für neue Zellen berücksichtigen und die Wahrscheinlichkeit für weitere Kombinationen berechnen!)
						else {
							$spielfeld->[$y_wert]->[$spalte] = 0;
							$zu_eliminierende_zellen->[$y_wert]->[$spalte] = 0;
						}
					}
				}
			}	
		}
	} 
	## Punktzahl für diesen Zug ans Hauptprogramm zurückgeben:
	return $points;
}
##

## 3-oder-mehr-Kombinationen finden
sub find_combos {
	
	## übergebene Variablen einlesen:
	my $spielfeld = shift;
	my ($x0, $x1, $y0, $y1) = @_;
	
	## Array aufbauen:
	my $zu_eliminierende_zellen; 
	for (0..9) {
		push @{ $zu_eliminierende_zellen } , [0,0,0,0,0,0,0,0,0,0];
	}
	
	### Kombos suchen:
	## Waagrecht:	
	foreach my $y ($y0..$y1) {
		foreach my $x ($x0-2..$x1) {
			
			next if $x < 0;	

			my %werte;
			my $is_bomb;
			
			foreach my $kombo_laenge (5,4,3) {
			
				next if $x > 10-$kombo_laenge; 														## Kombinationen nur möglich, wenn sich die Zelle nicht zu weit rechts befindet
				
				%werte = ();																		
				map {$werte{$_} = 1}  @{$spielfeld->[$y]}[$x..$x+$kombo_laenge-1] ;					## Alle Werte der 5/4/3 Zellen in Hash einlesen
				$is_bomb = 1;
				foreach my $wert (keys %werte) {								
					$is_bomb = 0 if $wert !~ m/[12345]/;											## Prüfen, ob (nur) Bomben vertreten sind
				}
				if (((scalar keys %werte == 1) or $is_bomb) and $spielfeld->[$y]->[$x] ne 0) {		## Falls nur ein Key im Hash vorhanden ist (also alle Zellen gleichen Inhalt haben) oder nur Bomben,...
					
					foreach my $zelle (@{$zu_eliminierende_zellen->[$y]}[$x..$x+$kombo_laenge-1]) {	## ...dann ist eine Kombination gefunden, die Zellen werden fürs Löschen vorgemerkt
						$zelle=1;
					}
					next;		
				}
			}
		}
	}
	
	## Senkrecht:
	foreach my $x ($x0..$x1) {
		
		foreach my $y ($y0-2..$y1) {
						
			next if $y < 0;	
			
			my %werte;
			my $is_bomb;
			
			foreach my $kombo_laenge (5,4,3) {
			
				next if $y > 10-$kombo_laenge; 	
				%werte = ();
				map {$werte{$_} = 1} (map { $spielfeld->[$_]->[$x] } $y..$y+$kombo_laenge-1) ;
				$is_bomb = 1;
				foreach my $wert (keys %werte) {
					$is_bomb = 0 if $wert !~ m/[12345]/;
				}
				if (((scalar keys %werte == 1) or $is_bomb) and $spielfeld->[$y]->[$x] ne 0) {
					
					foreach ($y..$y+$kombo_laenge-1) {
						$zu_eliminierende_zellen->[$_]->[$x] = 1
					}
					next;
				}
			}
		}
	}
	
	## Gefundene Zellen an andere Subfunktion (get_points) übergeben:
	return $zu_eliminierende_zellen;
		
}
##

